@@ -1,6 +1,6 @@
 use std::io::Write;
 
-use clap::*;
+use clap::{Arg, ArgAction, Command, ValueHint};
 
 use crate::generator::{utils, Generator};
 use crate::INTERNAL_ERROR_MSG;
@@ -11,10 +11,14 @@ pub struct Zsh;
 
 impl Generator for Zsh {
     fn file_name(&self, name: &str) -> String {
-        format!("_{}", name)
+        format!("_{name}")
     }
 
-    fn generate(&self, app: &App, buf: &mut dyn Write) {
+    fn generate(&self, cmd: &Command, buf: &mut dyn Write) {
+        let bin_name = cmd
+            .get_bin_name()
+            .expect("crate::generate should have set the bin_name");
+
         w!(
             buf,
             format!(
@@ -39,11 +43,16 @@ _{name}() {{
 
 {subcommand_details}
 
-_{name} \"$@\"",
-                name = app.get_bin_name().unwrap(),
-                initial_args = get_args_of(app, None),
-                subcommands = get_subcommands_of(app),
-                subcommand_details = subcommand_details(app)
+if [ \"$funcstack[1]\" = \"_{name}\" ]; then
+    _{name} \"$@\"
+else
+    compdef _{name} {name}
+fi
+",
+                name = bin_name,
+                initial_args = get_args_of(cmd, None),
+                subcommands = get_subcommands_of(cmd),
+                subcommand_details = subcommand_details(cmd)
             )
             .as_bytes()
         );
@@ -77,10 +86,12 @@ _{name} \"$@\"",
 //     )
 //     _describe -t commands 'rustup commands' commands "$@"
 //
-fn subcommand_details(p: &App) -> String {
+fn subcommand_details(p: &Command) -> String {
     debug!("subcommand_details");
 
-    let name = p.get_bin_name().unwrap();
+    let bin_name = p
+        .get_bin_name()
+        .expect("crate::generate should have set the bin_name");
 
     let mut ret = vec![];
 
@@ -92,8 +103,8 @@ _{bin_name_underscore}_commands() {{
     local commands; commands=({subcommands_and_args})
     _describe -t commands '{bin_name} commands' commands \"$@\"
 }}",
-        bin_name_underscore = name.replace(" ", "__"),
-        bin_name = name,
+        bin_name_underscore = bin_name.replace(' ', "__"),
+        bin_name = bin_name,
         subcommands_and_args = subcommands_of(p)
     );
     ret.push(parent_text);
@@ -104,8 +115,8 @@ _{bin_name_underscore}_commands() {{
     all_subcommands.sort();
     all_subcommands.dedup();
 
-    for &(_, ref bin_name) in &all_subcommands {
-        debug!("subcommand_details:iter: bin_name={}", bin_name);
+    for (_, ref bin_name) in &all_subcommands {
+        debug!("subcommand_details:iter: bin_name={bin_name}");
 
         ret.push(format!(
             "\
@@ -114,7 +125,7 @@ _{bin_name_underscore}_commands() {{
     local commands; commands=({subcommands_and_args})
     _describe -t commands '{bin_name} commands' commands \"$@\"
 }}",
-            bin_name_underscore = bin_name.replace(" ", "__"),
+            bin_name_underscore = bin_name.replace(' ', "__"),
             bin_name = bin_name,
             subcommands_and_args =
                 subcommands_of(parser_of(p, bin_name).expect(INTERNAL_ERROR_MSG))
@@ -135,23 +146,21 @@ _{bin_name_underscore}_commands() {{
 // A snippet from rustup:
 //         'show:Show the active and installed toolchains'
 //      'update:Update Rust toolchains'
-fn subcommands_of(p: &App) -> String {
+fn subcommands_of(p: &Command) -> String {
     debug!("subcommands_of");
 
     let mut segments = vec![];
 
-    fn add_subcommands(subcommand: &App, name: &str, ret: &mut Vec<String>) {
+    fn add_subcommands(subcommand: &Command, name: &str, ret: &mut Vec<String>) {
         debug!("add_subcommands");
 
         let text = format!(
             "'{name}:{help}' \\",
             name = name,
-            help = escape_help(subcommand.get_about().unwrap_or(""))
+            help = escape_help(&subcommand.get_about().unwrap_or_default().to_string())
         );
 
-        if !text.is_empty() {
-            ret.push(text);
-        }
+        ret.push(text);
     }
 
     // The subcommands
@@ -205,7 +214,7 @@ fn subcommands_of(p: &App) -> String {
 //    [name_hyphen] = The full space delineated bin_name, but replace spaces with hyphens
 //    [repeat] = From the same recursive calls, but for all subcommands
 //    [subcommand_args] = The same as zsh::get_args_of
-fn get_subcommands_of(parent: &App) -> String {
+fn get_subcommands_of(parent: &Command) -> String {
     debug!(
         "get_subcommands_of: Has subcommands...{:?}",
         parent.has_subcommands()
@@ -218,16 +227,14 @@ fn get_subcommands_of(parent: &App) -> String {
     let subcommand_names = utils::subcommands(parent);
     let mut all_subcommands = vec![];
 
-    for &(ref name, ref bin_name) in &subcommand_names {
+    for (ref name, ref bin_name) in &subcommand_names {
         debug!(
-            "get_subcommands_of:iter: parent={}, name={}, bin_name={}",
+            "get_subcommands_of:iter: parent={}, name={name}, bin_name={bin_name}",
             parent.get_name(),
-            name,
-            bin_name,
         );
-        let mut segments = vec![format!("({})", name)];
+        let mut segments = vec![format!("({name})")];
         let subcommand_args = get_args_of(
-            parser_of(parent, &*bin_name).expect(INTERNAL_ERROR_MSG),
+            parser_of(parent, bin_name).expect(INTERNAL_ERROR_MSG),
             Some(parent),
         );
 
@@ -236,7 +243,7 @@ fn get_subcommands_of(parent: &App) -> String {
         }
 
         // Get the help text of all child subcommands.
-        let children = get_subcommands_of(parser_of(parent, &*bin_name).expect(INTERNAL_ERROR_MSG));
+        let children = get_subcommands_of(parser_of(parent, bin_name).expect(INTERNAL_ERROR_MSG));
 
         if !children.is_empty() {
             segments.push(children);
@@ -246,6 +253,10 @@ fn get_subcommands_of(parent: &App) -> String {
         all_subcommands.push(segments.join("\n"));
     }
 
+    let parent_bin_name = parent
+        .get_bin_name()
+        .expect("crate::generate should have set the bin_name");
+
     format!(
         "
     case $state in
@@ -259,20 +270,20 @@ fn get_subcommands_of(parent: &App) -> String {
     ;;
 esac",
         name = parent.get_name(),
-        name_hyphen = parent.get_bin_name().unwrap().replace(" ", "-"),
+        name_hyphen = parent_bin_name.replace(' ', "-"),
         subcommands = all_subcommands.join("\n"),
         pos = parent.get_positionals().count() + 1
     )
 }
 
-// Get the App for a given subcommand tree.
+// Get the Command for a given subcommand tree.
 //
-// Given the bin_name "a b c" and the App for "a" this returns the "c" App.
-// Given the bin_name "a b c" and the App for "b" this returns the "c" App.
-fn parser_of<'help, 'app>(parent: &'app App<'help>, bin_name: &str) -> Option<&'app App<'help>> {
+// Given the bin_name "a b c" and the Command for "a" this returns the "c" Command.
+// Given the bin_name "a b c" and the Command for "b" this returns the "c" Command.
+fn parser_of<'cmd>(parent: &'cmd Command, bin_name: &str) -> Option<&'cmd Command> {
     debug!("parser_of: p={}, bin_name={}", parent.get_name(), bin_name);
 
-    if bin_name == parent.get_bin_name().unwrap_or(&String::new()) {
+    if bin_name == parent.get_bin_name().unwrap_or_default() {
         return Some(parent);
     }
 
@@ -305,7 +316,7 @@ fn parser_of<'help, 'app>(parent: &'app App<'help>, bin_name: &str) -> Option<&'
 //    -C: modify the $context internal variable
 //    -s: Allow stacking of short args (i.e. -a -b -c => -abc)
 //    -S: Do not complete anything after '--' and treat those as argument values
-fn get_args_of(parent: &App, p_global: Option<&App>) -> String {
+fn get_args_of(parent: &Command, p_global: Option<&Command>) -> String {
     debug!("get_args_of");
 
     let mut segments = vec![String::from("_arguments \"${_arguments_options[@]}\" \\")];
@@ -326,9 +337,12 @@ fn get_args_of(parent: &App, p_global: Option<&App>) -> String {
     }
 
     if parent.has_subcommands() {
+        let parent_bin_name = parent
+            .get_bin_name()
+            .expect("crate::generate should have set the bin_name");
         let subcommand_bin_name = format!(
             "\":: :_{name}_commands\" \\",
-            name = parent.get_bin_name().as_ref().unwrap().replace(" ", "__")
+            name = parent_bin_name.replace(' ', "__")
         );
         segments.push(subcommand_bin_name);
 
@@ -342,23 +356,24 @@ fn get_args_of(parent: &App, p_global: Option<&App>) -> String {
 
 // Uses either `possible_vals` or `value_hint` to give hints about possible argument values
 fn value_completion(arg: &Arg) -> Option<String> {
-    if let Some(values) = &arg.get_possible_values() {
+    if let Some(values) = utils::possible_values(arg) {
         if values
             .iter()
-            .any(|value| !value.is_hidden() && value.get_help().is_some())
+            .any(|value| !value.is_hide_set() && value.get_help().is_some())
         {
             Some(format!(
                 "(({}))",
                 values
                     .iter()
                     .filter_map(|value| {
-                        if value.is_hidden() {
+                        if value.is_hide_set() {
                             None
                         } else {
                             Some(format!(
                                 r#"{name}\:"{tooltip}""#,
                                 name = escape_value(value.get_name()),
-                                tooltip = value.get_help().map(escape_help).unwrap_or_default()
+                                tooltip =
+                                    escape_help(&value.get_help().unwrap_or_default().to_string()),
                             ))
                         }
                     })
@@ -370,7 +385,8 @@ fn value_completion(arg: &Arg) -> Option<String> {
                 "({})",
                 values
                     .iter()
-                    .filter_map(PossibleValue::get_visible_name)
+                    .filter(|pv| !pv.is_hide_set())
+                    .map(|n| n.get_name())
                     .collect::<Vec<_>>()
                     .join(" ")
             ))
@@ -406,34 +422,43 @@ fn value_completion(arg: &Arg) -> Option<String> {
 /// Escape help string inside single quotes and brackets
 fn escape_help(string: &str) -> String {
     string
-        .replace("\\", "\\\\")
-        .replace("'", "'\\''")
-        .replace("[", "\\[")
-        .replace("]", "\\]")
+        .replace('\\', "\\\\")
+        .replace('\'', "'\\''")
+        .replace('[', "\\[")
+        .replace(']', "\\]")
+        .replace(':', "\\:")
+        .replace('$', "\\$")
+        .replace('`', "\\`")
+        .replace('\n', " ")
 }
 
 /// Escape value string inside single quotes and parentheses
 fn escape_value(string: &str) -> String {
     string
-        .replace("\\", "\\\\")
-        .replace("'", "'\\''")
-        .replace("(", "\\(")
-        .replace(")", "\\)")
-        .replace(" ", "\\ ")
+        .replace('\\', "\\\\")
+        .replace('\'', "'\\''")
+        .replace('[', "\\[")
+        .replace(']', "\\]")
+        .replace(':', "\\:")
+        .replace('$', "\\$")
+        .replace('`', "\\`")
+        .replace('(', "\\(")
+        .replace(')', "\\)")
+        .replace(' ', "\\ ")
 }
 
-fn write_opts_of(p: &App, p_global: Option<&App>) -> String {
+fn write_opts_of(p: &Command, p_global: Option<&Command>) -> String {
     debug!("write_opts_of");
 
     let mut ret = vec![];
 
     for o in p.get_opts() {
-        debug!("write_opts_of:iter: o={}", o.get_name());
+        debug!("write_opts_of:iter: o={}", o.get_id());
 
-        let help = o.get_help().map_or(String::new(), escape_help);
+        let help = escape_help(&o.get_help().unwrap_or_default().to_string());
         let conflicts = arg_conflicts(p, o, p_global);
 
-        let multiple = if o.is_set(ArgSettings::MultipleOccurrences) {
+        let multiple = if let ArgAction::Count | ArgAction::Append = o.get_action() {
             "*"
         } else {
             ""
@@ -444,24 +469,14 @@ fn write_opts_of(p: &App, p_global: Option<&App>) -> String {
             Some(val) => val[0].to_string(),
         };
         let vc = match value_completion(o) {
-            Some(val) => format!(":{}:{}", vn, val),
-            None => format!(":{}: ", vn),
-        };
-        let vc = match o.get_num_vals() {
-            Some(num_vals) => vc.repeat(num_vals),
-            None => vc,
+            Some(val) => format!(":{vn}:{val}"),
+            None => format!(":{vn}: "),
         };
+        let vc = vc.repeat(o.get_num_args().expect("built").min_values());
 
         if let Some(shorts) = o.get_short_and_visible_aliases() {
             for short in shorts {
-                let s = format!(
-                    "'{conflicts}{multiple}-{arg}+[{help}]{value_completion}' \\",
-                    conflicts = conflicts,
-                    multiple = multiple,
-                    arg = short,
-                    value_completion = vc,
-                    help = help
-                );
+                let s = format!("'{conflicts}{multiple}-{short}+[{help}]{vc}' \\");
 
                 debug!("write_opts_of:iter: Wrote...{}", &*s);
                 ret.push(s);
@@ -469,14 +484,7 @@ fn write_opts_of(p: &App, p_global: Option<&App>) -> String {
         }
         if let Some(longs) = o.get_long_and_visible_aliases() {
             for long in longs {
-                let l = format!(
-                    "'{conflicts}{multiple}--{arg}=[{help}]{value_completion}' \\",
-                    conflicts = conflicts,
-                    multiple = multiple,
-                    arg = long,
-                    value_completion = vc,
-                    help = help
-                );
+                let l = format!("'{conflicts}{multiple}--{long}=[{help}]{vc}' \\");
 
                 debug!("write_opts_of:iter: Wrote...{}", &*l);
                 ret.push(l);
@@ -487,21 +495,21 @@ fn write_opts_of(p: &App, p_global: Option<&App>) -> String {
     ret.join("\n")
 }
 
-fn arg_conflicts(app: &App, arg: &Arg, app_global: Option<&App>) -> String {
+fn arg_conflicts(cmd: &Command, arg: &Arg, app_global: Option<&Command>) -> String {
     fn push_conflicts(conflicts: &[&Arg], res: &mut Vec<String>) {
         for conflict in conflicts {
             if let Some(s) = conflict.get_short() {
-                res.push(format!("-{}", s));
+                res.push(format!("-{s}"));
             }
 
             if let Some(l) = conflict.get_long() {
-                res.push(format!("--{}", l));
+                res.push(format!("--{l}"));
             }
         }
     }
 
     let mut res = vec![];
-    match (app_global, arg.get_global()) {
+    match (app_global, arg.is_global_set()) {
         (Some(x), true) => {
             let conflicts = x.get_arg_conflicts_with(arg);
 
@@ -512,7 +520,7 @@ fn arg_conflicts(app: &App, arg: &Arg, app_global: Option<&App>) -> String {
             push_conflicts(&conflicts, &mut res);
         }
         (_, _) => {
-            let conflicts = app.get_arg_conflicts_with(arg);
+            let conflicts = cmd.get_arg_conflicts_with(arg);
 
             if conflicts.is_empty() {
                 return String::new();
@@ -525,31 +533,25 @@ fn arg_conflicts(app: &App, arg: &Arg, app_global: Option<&App>) -> String {
     format!("({})", res.join(" "))
 }
 
-fn write_flags_of(p: &App, p_global: Option<&App>) -> String {
+fn write_flags_of(p: &Command, p_global: Option<&Command>) -> String {
     debug!("write_flags_of;");
 
     let mut ret = vec![];
 
     for f in utils::flags(p) {
-        debug!("write_flags_of:iter: f={}", f.get_name());
+        debug!("write_flags_of:iter: f={}", f.get_id());
 
-        let help = f.get_help().map_or(String::new(), escape_help);
+        let help = escape_help(&f.get_help().unwrap_or_default().to_string());
         let conflicts = arg_conflicts(p, &f, p_global);
 
-        let multiple = if f.is_set(ArgSettings::MultipleOccurrences) {
+        let multiple = if let ArgAction::Count | ArgAction::Append = f.get_action() {
             "*"
         } else {
             ""
         };
 
         if let Some(short) = f.get_short() {
-            let s = format!(
-                "'{conflicts}{multiple}-{arg}[{help}]' \\",
-                multiple = multiple,
-                conflicts = conflicts,
-                arg = short,
-                help = help
-            );
+            let s = format!("'{conflicts}{multiple}-{short}[{help}]' \\");
 
             debug!("write_flags_of:iter: Wrote...{}", &*s);
 
@@ -557,13 +559,7 @@ fn write_flags_of(p: &App, p_global: Option<&App>) -> String {
 
             if let Some(short_aliases) = f.get_visible_short_aliases() {
                 for alias in short_aliases {
-                    let s = format!(
-                        "'{conflicts}{multiple}-{arg}[{help}]' \\",
-                        multiple = multiple,
-                        conflicts = conflicts,
-                        arg = alias,
-                        help = help
-                    );
+                    let s = format!("'{conflicts}{multiple}-{alias}[{help}]' \\",);
 
                     debug!("write_flags_of:iter: Wrote...{}", &*s);
 
@@ -573,13 +569,7 @@ fn write_flags_of(p: &App, p_global: Option<&App>) -> String {
         }
 
         if let Some(long) = f.get_long() {
-            let l = format!(
-                "'{conflicts}{multiple}--{arg}[{help}]' \\",
-                conflicts = conflicts,
-                multiple = multiple,
-                arg = long,
-                help = help
-            );
+            let l = format!("'{conflicts}{multiple}--{long}[{help}]' \\");
 
             debug!("write_flags_of:iter: Wrote...{}", &*l);
 
@@ -587,13 +577,7 @@ fn write_flags_of(p: &App, p_global: Option<&App>) -> String {
 
             if let Some(aliases) = f.get_visible_aliases() {
                 for alias in aliases {
-                    let l = format!(
-                        "'{conflicts}{multiple}--{arg}[{help}]' \\",
-                        conflicts = conflicts,
-                        multiple = multiple,
-                        arg = alias,
-                        help = help
-                    );
+                    let l = format!("'{conflicts}{multiple}--{alias}[{help}]' \\");
 
                     debug!("write_flags_of:iter: Wrote...{}", &*l);
 
@@ -606,19 +590,55 @@ fn write_flags_of(p: &App, p_global: Option<&App>) -> String {
     ret.join("\n")
 }
 
-fn write_positionals_of(p: &App) -> String {
+fn write_positionals_of(p: &Command) -> String {
     debug!("write_positionals_of;");
 
     let mut ret = vec![];
 
-    for arg in p.get_positionals() {
-        debug!("write_positionals_of:iter: arg={}", arg.get_name());
+    // Completions for commands that end with two Vec arguments require special care.
+    // - You can have two Vec args separated with a custom value terminator.
+    // - You can have two Vec args with the second one set to last (raw sets last)
+    //   which will require a '--' separator to be used before the second argument
+    //   on the command-line.
+    //
+    // We use the '-S' _arguments option to disable completion after '--'. Thus, the
+    // completion for the second argument in scenario (B) does not need to be emitted
+    // because it is implicitly handled by the '-S' option.
+    // We only need to emit the first catch-all.
+    //
+    // Have we already emitted a catch-all multi-valued positional argument
+    // without a custom value terminator?
+    let mut catch_all_emitted = false;
 
-        let cardinality = if arg.is_set(ArgSettings::MultipleValues)
-            || arg.is_set(ArgSettings::MultipleOccurrences)
-        {
-            "*:"
-        } else if !arg.is_set(ArgSettings::Required) {
+    for arg in p.get_positionals() {
+        debug!("write_positionals_of:iter: arg={}", arg.get_id());
+
+        let num_args = arg.get_num_args().expect("built");
+        let is_multi_valued = num_args.max_values() > 1;
+
+        if catch_all_emitted && (arg.is_last_set() || is_multi_valued) {
+            // This is the final argument and it also takes multiple arguments.
+            // We've already emitted a catch-all positional argument so we don't need
+            // to emit anything for this argument because it is implicitly handled by
+            // the use of the '-S' _arguments option.
+            continue;
+        }
+
+        let cardinality_value;
+        // If we have any subcommands, we'll emit a catch-all argument, so we shouldn't
+        // emit one here.
+        let cardinality = if is_multi_valued && !p.has_subcommands() {
+            match arg.get_value_terminator() {
+                Some(terminator) => {
+                    cardinality_value = format!("*{}:", escape_value(terminator));
+                    cardinality_value.as_str()
+                }
+                None => {
+                    catch_all_emitted = true;
+                    "*:"
+                }
+            }
+        } else if !arg.is_required_set() {
             ":"
         } else {
             ""
@@ -627,20 +647,46 @@ fn write_positionals_of(p: &App) -> String {
         let a = format!(
             "'{cardinality}:{name}{help}:{value_completion}' \\",
             cardinality = cardinality,
-            name = arg.get_name(),
+            name = arg.get_id(),
             help = arg
                 .get_help()
-                .map_or("".to_owned(), |v| " -- ".to_owned() + v)
-                .replace("[", "\\[")
-                .replace("]", "\\]")
-                .replace(":", "\\:"),
-            value_completion = value_completion(arg).unwrap_or_else(|| "".to_string())
+                .map(|s| s.to_string())
+                .map(|v| " -- ".to_owned() + &v)
+                .unwrap_or_else(|| "".to_owned())
+                .replace('[', "\\[")
+                .replace(']', "\\]")
+                .replace('\'', "'\\''")
+                .replace(':', "\\:"),
+            value_completion = value_completion(arg).unwrap_or_default()
         );
 
-        debug!("write_positionals_of:iter: Wrote...{}", a);
+        debug!("write_positionals_of:iter: Wrote...{a}");
 
         ret.push(a);
     }
 
     ret.join("\n")
 }
+
+#[cfg(test)]
+mod tests {
+    use crate::shells::zsh::{escape_help, escape_value};
+
+    #[test]
+    fn test_escape_value() {
+        let raw_string = "\\ [foo]() `bar https://$PATH";
+        assert_eq!(
+            escape_value(raw_string),
+            "\\\\\\ \\[foo\\]\\(\\)\\ \\`bar\\ https\\://\\$PATH"
+        );
+    }
+
+    #[test]
+    fn test_escape_help() {
+        let raw_string = "\\ [foo]() `bar https://$PATH";
+        assert_eq!(
+            escape_help(raw_string),
+            "\\\\ \\[foo\\]() \\`bar https\\://\\$PATH"
+        );
+    }
+}
